#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "mem_alloc.h"

#define MAX_SIZE 13000000
#define MIN_SIZE 100000

struct buffer * alloc_buffer (unsigned int);
	
/* function who permit to allocate one (struct buffer *) element 
 * with a buffer from size bytes */
struct buffer * alloc_buffer (unsigned int size)
{
	struct buffer *buff;

	if (size == 0)
	{
		return (NULL);
	}
	if ((buff = (struct buffer *) 
			malloc (sizeof (struct buffer))) == NULL)
		return (NULL);
	
	if ((buff->data = (char *) malloc (size)) == NULL)
	{
		free (buff);
		return (NULL);
	}	
	
	buff->size = size;
	buff->next = NULL;
	return (buff);
}

/* function who permit to allocate size bytes of memory in a 
 * no MMU sytem by allocating lots of small amount of memory...
 * this memory can be use with a struct buffer * chain list... 
 * buff become the list head... 
 * this function use a recursive process... she try to allocate 
 * the maximum of adjacent memory in one struct buff element 
 * (via the malloc sytem call)...
 * and then, she call herself to finish the chain list buid */
struct buffer * alloc_mem (unsigned int size, unsigned int *get_size)
{
	unsigned int elt_size = size;
	struct buffer *buff;
	
	buff = NULL;

	if (size == 0)
		return (NULL);

	while ((buff = alloc_buffer (elt_size)) == NULL)
	{
		fprintf (stderr, "echec to allocate %d bytes : %d\n", elt_size);
		
		if (elt_size < MIN_SIZE)	
			return (buff);
		
		elt_size = (unsigned int) ((elt_size / 10) * 9);
	}	
	fprintf (stderr, "succes to allocate %d bytes\n", elt_size);
	buff->next = alloc_mem (size - elt_size, get_size);
	*get_size += elt_size;

	return (buff);
}

/* function who free the memory of all the element of a struct buffer 
 * chain list */
void free_mem (struct buffer *buff)
{
	struct buffer *tmp;
	
	while (buff != NULL)
	{
		fprintf (stderr, "size = %d\n", buff->size);
		tmp = buff;
		buff = buff->next;
		free (tmp->data);
		free (tmp);
	}

	return;
}
/*
int main (int argc, char **argv)
{
	unsigned int size, get_size;
	unsigned int curr_size = 0;
	struct buffer *head;
	
	if (argc != 2)
	{	
		fprintf (stderr, "bad parameter number\n");
		return (-1);
	}
	size = atoi (argv[1]);
	if (!(size < MAX_SIZE))
	{
		fprintf (stderr, "can't allocate %d bytes (max = %d)\n", 
				size, MAX_SIZE);
		return (-1);
	}
	head = alloc_mem (size, &get_size);
	fprintf (stdout, "get %d bytes for a %d bytes request\n", 
			get_size, size);
	free_mem (head);
	return (0);
}

int main ()
{
	int desc, nb_read, i, size;
	unsigned short *buffer, *ptr, result;

	if ((desc = open ("/dev/ppp", O_RDONLY)) == -1)
	{
		perror ("open ()");
		return (-1);
	}
	if ((buffer = (unsigned short *) malloc 
			(BUFFER_SIZE * sizeof (unsigned short))) == NULL)
	{
		perror ("malloc ()");
		return (-1);
	}

	size = BUFFER_SIZE * sizeof (unsigned short);
	ptr = buffer;
	while (size > 0)
	{
		if ((nb_read = read (desc, (unsigned char *) ptr, size)) == -1)
		{
			perror ("read ()");
			return (-1);
		}
		size -= nb_read;
		((unsigned char *) ptr) += nb_read;
	}
	for (i = 0; i < BUFFER_SIZE; i++)
	{
		fprintf (stdout, "%d\n", buffer[i]);
	}
	free (buffer);
	close (desc);
	return (0);
}	
*/
