#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/semaphore.h>
#include <linux/slab.h>
//#include <asm-m68knommu/m5282sim.h>
#include <asm/coldfire.h>
#include "int.h"
#include "qadc_cmd.h"

#define QADCMCR  (MCF_IPSBAR + 0x00190000)  /* 16 bit */
#define QADCTEST (MCF_IPSBAR + 0x00190002)  /* 16 bit */
#define PORTQA   (MCF_IPSBAR + 0x00190006)  /* 8 bit */
#define PORTQB   (MCF_IPSBAR + 0x00190007)  /* 8 bit */
#define DDRQA    (MCF_IPSBAR + 0x00190008)  /* 8 bit */
#define DDRQB    (MCF_IPSBAR + 0x00190009)  /* 8 bit */
#define QACR0    (MCF_IPSBAR + 0x0019000A)  /* 16 bit */
#define QACR1    (MCF_IPSBAR + 0x0019000C)  /* 16 bit */
#define QACR2    (MCF_IPSBAR + 0x0019000E)  /* 16 bit */
#define QASR0    (MCF_IPSBAR + 0x00190010)  /* 16 bit */
#define QASR1    (MCF_IPSBAR + 0x00190012)  /* 16 bit */
#define CCW      (MCF_IPSBAR + 0x00190200)  /* N bit */
#define RJURR    (MCF_IPSBAR + 0x00190280)  /* N bit */
#define LJSRR    (MCF_IPSBAR + 0x00190300)  /* N bit */
#define LJURR    (MCF_IPSBAR + 0x00190380)  /* N bit */

#define GPTAPORT (MCF_IPSBAR + 0x001A001D)  /* 8 bit */
#define GPTADDR  (MCF_IPSBAR + 0x001A001E)  /* 8 bit */

#define CCW_NB 64 /* number of entries in the CCW table */

#define COMPLETE_Q1_IRQ 37
#define COMPLETE_Q2_IRQ 38
#define PAUSE_Q1_IRQ 39
#define PAUSE_Q2_IRQ 40

/* configuration range for the Input Sample Time Register */
#define DFLT_IST 1
#define MIN_IST 0
#define MAX_IST 3

/* global value for the QADC clock prescaler */
#define DFLT_QPR 0x20
#define QPR_MIN 0x00
#define QPR_MAX 0x7f

/* global value for the QADC timer */
#define NO_TIMER 0x11 /* continuous scan mode */
#define TIMER_MIN 0x14
#define TIMER_MAX 0x1e

#define NB_BUFFER 2
#define BUFFER_SIZE 1000


/* Use our own dbg macro 
 * and enable debug by default */
#ifdef DEBUG    
#define dbg(format, arg...) printk(KERN_EMERG __FILE__ ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) 
#endif
	
/* Version Information */
#define DRIVER_AUTHOR "Simon Guinot"
#define DRIVER_DESC "driver to use the QADC on the 5282 motorola board"

/* enable dynamic major allocation by default... 
 * but user can manualy fix it by passing a parameter during 
 * the insmod process */
static int qadc_major = 108;

MODULE_PARM(qadc_major, "i");
MODULE_PARM_DESC(qadc_major, "to fix manualy major number");

struct buffer
{
	unsigned short data[BUFFER_SIZE];
	unsigned short *data_start;
	unsigned short size;
};

struct qadc
{
	unsigned int open_count;
	struct buffer buff[NB_BUFFER];
	struct buffer *write_ptr;
	struct buffer *read_ptr;
	struct semaphore qadc_sem; /* to protect this struct */
	wait_queue_head_t qadc_queue;
	struct irq_info *qadc_irq_info;
	unsigned char qadc_timer;
	unsigned char qadc_ist; /* input sample time configuration */
	unsigned char qadc_qpr; /* prescaler clock divider */
	struct input_chan in_chan;
};	

static void start_qadc (struct qadc *);
static void stop_qadc (void);
static void qadc_irqhandler (int, void *, struct pt_regs *);
static int qadc_open (struct inode *, struct file *);
static ssize_t qadc_read (struct file *, char *, size_t, loff_t *);
static int qadc_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
static int qadc_release (struct inode *, struct file *);
static int __init qadc_init (void);
static void __exit qadc_exit (void);

static struct file_operations qadc_fops = 
{
	owner:		THIS_MODULE,
	read:		qadc_read,
	open:		qadc_open,
	ioctl:		qadc_ioctl,
	release:	qadc_release,
};

static void qadc_irqhandler (int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
	struct qadc *dev = (struct qadc *) dev_id;
	volatile unsigned short *qasr0 = (volatile unsigned short *) QASR0;

	for (i = 0; i < dev->in_chan.active_chan_nb; i++)
	{
		dev->write_ptr->data[dev->write_ptr->size++%BUFFER_SIZE] = 
			*((unsigned short *) (RJURR + i)) & 0x03ff;
	}
	dev->write_ptr->size = dev->write_ptr->size%BUFFER_SIZE;
	/*
	dev->write_ptr->size = (dev->write_ptr->size + 
			dev->in_chan.active_chan_nb)%BUFFER_SIZE;
	*/
	if (dev->write_ptr->size > (BUFFER_SIZE / 2))
		wake_up_interruptible (&dev->qadc_queue);
	
	*qasr0 &= 0x7fff;

	return;
/*
	switch (irq)
	{
		case (COMPLETE_Q1_IRQ + INTC0_VECT_BASE) :
			
			dbg("%s - end of queue 1\n", __FUNCTION__);
			
*/			/* clean the interrupt (the CF1 bit) */
/*			*qasr &= 0x7fff;
			
			break;
			
		case (COMPLETE_Q2_IRQ + INTC0_VECT_BASE) :
			
			dbg("%s - end of queue 2 => problem\n", __FUNCTION__);			
*/			/* clean the interrupt (the CF2 bit) */
//			*qasr &= 0xdfff;
			
			/* reenable queue 2 by sending a trigger */
//			*((volatile unsigned short *) QACR2) |= 0x2000;
			
/*			break;
			
		case (PAUSE_Q1_IRQ + INTC0_VECT_BASE) :
			
			dbg("%s - queue 1 paused => problem\n", __FUNCTION__);
			
*/			/* clean the interrupt (the PF1 bit) */
/*			*qasr &= 0xbfff;
			
			break;
			
		case (PAUSE_Q2_IRQ + INTC0_VECT_BASE) :
			
			dbg("%s - queue 2 paused\n", __FUNCTION__);
			
*/			/*get the position where queue 2 is paused in 
			 * the CCW table */
/*			ccw_pos = (*((volatile unsigned char *) QASR1) & 0x3f);
			
*/			/* clean the interrupt (the PF2 bit) */
/*			*qasr &= 0xefff;
			
			break;

		default :
			
			dbg("%s - PROBLEM :)\n", __FUNCTION__);
			break;
	}
	return;*/
}

/* function to enable stop mode for the qadc */
static void stop_qadc (void)
{
	dbg("enter %s", __FUNCTION__);
	*((volatile unsigned short *) QADCMCR) = 0x8000;
	return;
}

/* function who initiate the qadc */
static void start_qadc (struct qadc *dev)
{
	int i;
	
	volatile unsigned short *ccw = (volatile unsigned short *) (CCW);

	dbg("enter %s", __FUNCTION__);
	
	/* QADC register initialisation */
	*((volatile unsigned short *) QADCMCR) = 0x0000; /* control register */
	*((volatile unsigned short *) DDRQA) &= 0xff00; /* input mode for QA ports */
	if (dev->qadc_qpr > QPR_MAX )
	{
		dbg("bad prescaler clock diviser... fix it to default value");
		dev->qadc_qpr = DFLT_QPR;
	}
	
	*((volatile unsigned char *) QACR0) = dev->qadc_qpr; /* set the clock */

	if (dev->qadc_ist > MAX_IST)
	{
		dbg("bad input sample timer... fix it to default value");
		dev->qadc_ist = DFLT_IST;
	}
	/* construct the CCW table... we set PQA0 or AN52 
	 * as input channel */
	for (i = 0; i < dev->in_chan.active_chan_nb; i++)
	{
		ccw[i] = (dev->qadc_ist << 6) | dev->in_chan.chan[i];
	}
	ccw[i] = 63;
	/* enable periodic timer continuous scan mode for queue 1 and 
	 * associate an interruption to the queue 1 end */
	
	if ((!(dev->qadc_timer <= TIMER_MAX)) && 
			(!(dev->qadc_timer >= TIMER_MIN)))
	{
		dbg("bad timer value... use the no timer configuration");
		dev->qadc_timer = NO_TIMER;
	}
	*((volatile unsigned short *) QACR1) = ((1 << 15) | 
			(dev->qadc_timer << 8));

	/* stop the queue 1 after the ccw[0] convertion 
	 * the method is to set the queue 2 begining at the ccw[1] entry 
	 * and disable conversion for the queue 2 */
//	*((volatile unsigned short *) QACR2) = 0x0001;
//	*((volatile unsigned short *) QACR2) = 0x0000;
	
	/* send a software trigger to start conversion */
	*((volatile unsigned short *) QACR1) |= 0x2000;

	return;
}

/* method to open the qadc device */
static int qadc_open (struct inode *inode, struct file *file)
{
	struct qadc *dev = NULL;
	int retval = 0, i;
	struct irq_info *qadc_irq_info = NULL;
	
	dev = (struct qadc *) file->private_data;

	dbg("enter %s", __FUNCTION__);
	
	/* just for compatibility with older kernel version */
	MOD_INC_USE_COUNT;

	/* if it is the first open, it is time to 
	 * initiate some structure */
	if (!dev)
	{
		dbg("%s - alloc device", __FUNCTION__);
		dev = (struct qadc *) kmalloc 
			(sizeof (struct qadc), GFP_KERNEL);		
		qadc_irq_info = (struct irq_info *) kmalloc 
			(sizeof (struct irq_info), GFP_KERNEL);		
		if ((dev == NULL) || (qadc_irq_info == NULL))
		{
			retval = -ENOMEM;
			goto exit;
		}	
		memset (dev, 0x00, sizeof (struct qadc));
		for (i = 0; i < NB_BUFFER; i++)
		{
			dev->buff[i].size = 0;
		}
		dev->write_ptr = &dev->buff[0];
		dev->read_ptr = &dev->buff[1];
		dev->write_ptr->data_start = &dev->write_ptr->data[0];
		dev->read_ptr->data_start = &dev->read_ptr->data[0];
		dev->qadc_timer = TIMER_MIN;
		dev->qadc_qpr = DFLT_QPR;
		dev->qadc_ist = MAX_IST;
		dev->in_chan.active_chan_nb = 1; /* use only 1 channel by default */
		dev->in_chan.chan[0] = AN52; /* convert data frome the AN52 channel by default */
		init_waitqueue_head (&dev->qadc_queue);
		sema_init (&dev->qadc_sem, 1);
		dev->open_count = 0;
		file->private_data = dev;
	}

	/* lock this device */
	if (down_interruptible (&dev->qadc_sem))
	{	
		retval = -ERESTARTSYS;
		goto exit_free;
	}	

	/* initiate the qadc and enable the corresponding 
	 * interruption */
	if (dev->open_count == 0)
	{
		stop_qadc ();
		/* initiate the structure to request our interrupt vector */
		sprintf (qadc_irq_info->name, "adc_int");
		qadc_irq_info->vector = COMPLETE_Q1_IRQ + INTC0_VECT_MIN;
		qadc_irq_info->level = 0x37; /* level 6 and pririty 7 */
		qadc_irq_info->handler = qadc_irqhandler;
		qadc_irq_info->dev_id = (void *) dev;
		if ((retval = enable_interrupt_vector (qadc_irq_info)))
		{
			dbg("can't enable qadc interrupt vector");
			goto exit_sem;
		}	
		dev->qadc_irq_info = qadc_irq_info;
		start_qadc (dev);
	}
		
	/* increment our own counter */
	++dev->open_count;
	
	dbg("%s - open_count = %d", __FUNCTION__, dev->open_count);

exit_sem :
	up (&dev->qadc_sem);
	
exit_free :
	if (retval)
		kfree (dev);
	
exit :
	if (retval)
		MOD_DEC_USE_COUNT;

	return (retval);
}

/* method read () to transmit to a user, data resulting from analogic-numeric 
 * conversion */
static ssize_t qadc_read (struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct qadc *dev = (struct qadc *) file->private_data;
	struct buffer *tmp;
	unsigned short size = 0;

	if (!dev)
		return (-ENODEV);

	if (dev->read_ptr->size)
	{
		goto exit_copy_to;
	}

	dev->read_ptr->data_start = dev->read_ptr->data;
	tmp = dev->read_ptr;
	dev->read_ptr = dev->write_ptr;

	/* sleep until a interruption send a wake up signal */ 
	if (wait_event_interruptible (dev->qadc_queue, 
				(dev->read_ptr->size > (BUFFER_SIZE / 2))))
		return (-ERESTARTSYS);

	dev->write_ptr = tmp;
	
exit_copy_to :
	
	size = ((dev->read_ptr->size * sizeof (unsigned short)) < count)? 
		(dev->read_ptr->size * sizeof (unsigned short)) : count;
	
	__copy_to_user (buffer, (unsigned char *) dev->read_ptr->data_start, size);
	dev->read_ptr->data_start += (size / sizeof (unsigned short));	
	dev->read_ptr->size -= (size / sizeof (unsigned short));

	return (size);
}

static int qadc_ioctl (struct inode *inode, struct file *file, 
		unsigned int cmd, unsigned long arg)
{
        struct qadc *dev;
	struct input_chan in_chan_cmd;
        int retval = -ENOTTY;

        dev = (struct qadc *)file->private_data;

        /* lock this object */
        down (&dev->qadc_sem);

        dbg("%s - cmd 0x%.4x, arg %ld", __FUNCTION__, cmd, arg);

	switch (cmd)
        {
		case CMD_QPR :
			dev->qadc_qpr = arg;
			break;

		case CMD_IST :
			dev->qadc_ist = arg;
			break;

		case CMD_TIMER :
			dev->qadc_timer = arg;
			break;

		case CMD_CHAN :
			/* must check return value */
			copy_from_user (&in_chan_cmd, (void *) arg, 
					sizeof (struct input_chan));
			/* must check the user channel command */
			memcpy (&dev->in_chan, &in_chan_cmd, 
					sizeof (struct input_chan));
			break;

		default :
			dbg("%s - unrecognized command", __FUNCTION__);
			goto exit;
			break;
        }

	stop_qadc ();
	start_qadc (dev);
	retval = 0;

exit :
        /* unlock the device */
        up (&dev->qadc_sem);

        return retval;
}

/* the realease method is call when user want close () the device */
static int qadc_release (struct inode *inode, struct file *file)
{
	struct qadc *dev = NULL;
	int retval = 0;

	dbg("enter %s", __FUNCTION__);
	
	dev = (struct qadc *) file->private_data;

	stop_qadc ();
	
	/* lock this device */
	if (down_interruptible (&dev->qadc_sem))
	{	
		retval = -ERESTARTSYS;
		goto exit;
	}	

	/* if they is no user for this device... 
	 * we can free the irq and the adc structure */
	if ((--dev->open_count) == 0)
	{
		disable_interrupt_vector (dev->qadc_irq_info);
		up (&dev->qadc_sem);
		kfree (dev);
	}	

	MOD_DEC_USE_COUNT;

exit:
	return (retval);
}

static int __init qadc_init (void)
{
	dbg("enter %s", __FUNCTION__);
	
	/* create our qadc device */
	if (register_chrdev (qadc_major, "ppp", &qadc_fops) < 0)
	{
		dbg("can't get a major number");
		return (-1);
	}
	
	dbg("%s - qadc_major %d", __FUNCTION__, qadc_major);
	dbg("%s - init complete", __FUNCTION__);
	
	return (0);
}

static void __exit qadc_exit (void)
{
	dbg("enter %s", __FUNCTION__);

	if (unregister_chrdev (qadc_major, "ppp"))
		dbg("qadc : oops... can't release major number");	

	dbg("%s - exit ok", __FUNCTION__);
}

module_init (qadc_init);
module_exit (qadc_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
